/* authorize depending on yes/no response of external server */
/* See http://www.hoboes.com/Mimsy/?ART=639 for tutorial */
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"

#include "apr_strings.h"

/* create a request to send client information to an authentication server */
static char *makeAuthRequest(request_rec *request) {
	char *authRequest;
	char *remote_ip;
	char *uri;
	
	remote_ip = request->connection->remote_ip;
	uri = request->uri;
	
	authRequest = apr_pstrcat(request->pool, "GET /authorization.php?ip=", remote_ip, "&page=", uri, "\n", NULL);
	
	return authRequest;
}

static int logError(apr_status_t status, char *message, request_rec *request) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, status, request, "external_auth: error %s", message);
	return HTTP_SERVICE_UNAVAILABLE;
}

static int authenticate(request_rec *request, char *authRequest, char response[]) {
	apr_socket_t *sock;
	apr_sockaddr_t *sockaddr;
	apr_status_t status;
	//timeout appears to be in microseconds, so time out after 5 seconds
	apr_interval_time_t timeout = 5000000;
	apr_size_t item_size;

	if ((status = apr_sockaddr_info_get(&sockaddr, "216.92.252.156", APR_INET, 80, 0, request->pool)) != APR_SUCCESS) {
		return logError(status, "creating socket address", request);
	}

	if ((status = apr_socket_create(&sock, sockaddr->family, SOCK_STREAM, APR_PROTO_TCP, request->pool)) != APR_SUCCESS) {
		return logError(status, "creating socket", request);
	}
		
	if ((status = apr_socket_timeout_set(sock, timeout)) != APR_SUCCESS) {
		return logError(status, "setting socket timeout", request);
	}

	if ((status = apr_socket_connect(sock, sockaddr)) != APR_SUCCESS) {
		return logError(status, "connecting", request);
	}
	
	item_size = strlen(authRequest);
	if ((status = apr_socket_send(sock, authRequest, &item_size)) != APR_SUCCESS) {
		return logError(status, "sending query", request);
	}
	
	item_size = sizeof(response);
	if ((status = apr_socket_recv(sock, response, &item_size)) != APR_SUCCESS) {
		return logError(status, "receiving response", request);
	}

	apr_socket_close(sock);

	/* make sure it ends, and chop off carriage return */
	response[item_size] = '\0';
	if (item_size > 0) {
		if (response[item_size-1] == '\n') {
			response[item_size-1] = '\0';
		}
	}
	
	return APR_SUCCESS;
}

static int externalAuthorizer(request_rec *request) {
	const char *current_auth;
	char *authRequest;
	/* more characters than we will ever need for the auth server's response */
	char response[5000];
	int status;

	//if this is not a HOBOES auth type, bail and let someone else handle it
	current_auth = ap_auth_type(request);
	if (!current_auth || strcasecmp(current_auth, "hoboes")) {
		return DECLINED;
	}

	//create authorization request
	authRequest = makeAuthRequest(request);

	//make call to authentication server
	if ((status = authenticate(request, authRequest, response)) != APR_SUCCESS) {
		return status;
	}

	//parse authentication response
	if (strcmp("No", response) == 0) {
		return HTTP_FORBIDDEN;
	} else if (strcmp("Yes", response) == 0) {
		return OK;
	} else {
		return HTTP_SERVICE_UNAVAILABLE;
	}
}

static void myRegisterHooks(apr_pool_t *p) {
	ap_hook_check_user_id(externalAuthorizer, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA external_auth_module = {
	STANDARD20_MODULE_STUFF,
	NULL, /* create per-directory config structures */
	NULL, /* merge per-directory config structures */
	NULL, /* create per-server config structures */
	NULL, /* merge per-server config structures */
	NULL, /* command handlers */
	myRegisterHooks /* register hooks */
};